// ****************************************************************************
//             Copyright (C) Oasys Design Systems, Inc. 2004 - 2008
//                             All rights reserved.
// ****************************************************************************
#include "LLoad.h"
#include "LibLoadMsg.h"
#include "attr_enum.h"
#include "LibParam.h"

void
LLoad::processLibAttr(const LibGroup* llib)
{
  LibAttr* attr;
  LibAttrIter aiter;
  TLib& tlib = *d_lib.tlib();
  LibErrorT err;
  const char *libName = llib->firstName();

  if ((attr = llib->findAttr(K_capacitive_load_unit, err))) {
    LibValIter viter = attr->valueIter();
    LibVal* v = viter++;
    double capUnit = v->double_val();
    v = viter++;
    const char* capUnitStr = v->string_val();
    tlib.capUnitConversion(capUnit, capUnitStr);
  }
  if ((attr = llib->findAttr(K_time_unit, err))) {
    tlib.timeUnitConversion(attr->value()->string_val());
  }
  TLibScaleFactors* sf = tlib.defaultScaleFactors();
  TLibOpCnd&   nomOC = tlib.nominalOpCnd();
  aiter = llib->attrIter();
  while ((attr = aiter++)) {
    switch ((attr_enum)attr->typeID()) {
    case LIBERTY_ATTRENUM_leakage_power_unit                     : // enum( 1mW, 10pW, 10nW, 100uW, 100pW, 100nW, 1uW, 10uW, 1pW, or 1nW );
      tlib.leakagePowerUnitConversion(attr->value()->string_val());
      break;
    case LIBERTY_ATTRENUM_default_fanout_load                    : // float( > 0.0 ) := 1.0;
      tlib.defaultFanoutLoad(attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_default_max_capacitance                : // float( > 0.0 );
      tlib.defaultMaxCap(tlib.convertCap(attr->value()->double_val()));
      break;
    case LIBERTY_ATTRENUM_default_max_fanout                     : // float( > 0.0 );
      tlib.defaultMaxFanout(attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_default_max_transition                 : // float( > 0.0 );
      {
	double s = attr->value()->double_val();
	TVal maxSlew = tlib.convertTime(s);
	if (maxSlew > TSlew::max()) {
	  LIBLOADMSG(maxSlewTooBig, s);
	  tlib.defaultMaxSlew(TSlew::max());
	} else {
	  tlib.defaultMaxSlew(maxSlew);
	}
	break;
      }
    case LIBERTY_ATTRENUM_default_inout_pin_cap                  : // float( > 0.0 ) := 1.0;
      tlib.defaultInOutPinCap(tlib.convertCap(attr->value()->double_val()));
      break;
    case LIBERTY_ATTRENUM_default_input_pin_cap                  : // float( > 0.0 ) := 1.0;
      tlib.defaultInPinCap(tlib.convertCap(attr->value()->double_val()));
      break;
    case LIBERTY_ATTRENUM_default_output_pin_cap                 : // float( > 0.0 ) := 0.0;
      tlib.defaultOutPinCap(tlib.convertCap(attr->value()->double_val()));
      break;
    case LIBERTY_ATTRENUM_pulling_resistance_unit                : // enum( 1ohm, 100ohm, 10ohm, or 1kohm );
      tlib.resUnitConversion(attr->value()->string_val());
      break;
    case LIBERTY_ATTRENUM_time_unit                              : // enum( 1ps, 1ns, 10ps, or 100ps );
      // processed already
      break;
    case LIBERTY_ATTRENUM_capacitive_load_unit                   : // function( float, string );
      // processed already
      break;
    case LIBERTY_ATTRENUM_k_process_cell_fall                    : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_cell(TLKFD_process, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_cell_leakage_power           : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_cell_leakage_power(TLKFD_process, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_cell_rise                    : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_cell(TLKFD_process, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_drive_fall                   : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_drive(TLKFD_process, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_drive_rise                   : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_drive(TLKFD_process, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_fall_propagation             : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_propagation(TLKFD_process, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_fall_transition              : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_transition(TLKFD_process, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_hold_fall                    : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_hold(TLKFD_process, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_hold_rise                    : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_hold(TLKFD_process, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_internal_power               : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_internal_power(TLKFD_process, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_intrinsic_fall               : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_intrinsic(TLKFD_process, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_intrinsic_rise               : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_intrinsic(TLKFD_process, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_pin_cap                      : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_pin_cap(TLKFD_process, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_recovery_fall                : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_recovery(TLKFD_process, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_recovery_rise                : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_recovery(TLKFD_process, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_removal_fall                 : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_removal(TLKFD_process, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_removal_rise                 : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_removal(TLKFD_process, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_rise_propagation             : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_propagation(TLKFD_process, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_rise_transition              : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_transition(TLKFD_process, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_setup_fall                   : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_setup(TLKFD_process, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_setup_rise                   : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_setup(TLKFD_process, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_slope_fall                   : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_slope(TLKFD_process, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_slope_rise                   : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_slope(TLKFD_process, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_cell_fall                       : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_cell(TLKFD_temperature, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_cell_leakage_power              : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_cell_leakage_power(TLKFD_temperature, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_cell_rise                       : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_cell(TLKFD_temperature, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_drive_fall                      : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_drive(TLKFD_temperature, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_drive_rise                      : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_drive(TLKFD_temperature, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_fall_propagation                : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_propagation(TLKFD_temperature, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_fall_transition                 : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_transition(TLKFD_temperature, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_hold_fall                       : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_hold(TLKFD_temperature, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_hold_rise                       : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_hold(TLKFD_temperature, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_internal_power                  : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_internal_power(TLKFD_temperature, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_intrinsic_fall                  : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_intrinsic(TLKFD_temperature, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_intrinsic_rise                  : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_intrinsic(TLKFD_temperature, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_pin_cap                         : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_pin_cap(TLKFD_temperature, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_recovery_fall                   : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_recovery(TLKFD_temperature, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_recovery_rise                   : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_recovery(TLKFD_temperature, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_removal_fall                    : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_removal(TLKFD_temperature, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_removal_rise                    : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_removal(TLKFD_temperature, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_rise_propagation                : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_propagation(TLKFD_temperature, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_rise_transition                 : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_transition(TLKFD_temperature, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_setup_fall                      : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_setup(TLKFD_temperature, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_setup_rise                      : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_setup(TLKFD_temperature, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_slope_fall                      : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_slope(TLKFD_temperature, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_slope_rise                      : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_slope(TLKFD_temperature, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_cell_fall                       : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_cell(TLKFD_voltage, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_cell_leakage_power              : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_cell_leakage_power(TLKFD_voltage, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_cell_rise                       : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_cell(TLKFD_voltage, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_drive_fall                      : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_drive(TLKFD_voltage, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_drive_rise                      : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_drive(TLKFD_voltage, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_fall_propagation                : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_propagation(TLKFD_voltage, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_fall_transition                 : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_transition(TLKFD_voltage, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_hold_fall                       : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_hold(TLKFD_voltage, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_hold_rise                       : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_hold(TLKFD_voltage, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_internal_power                  : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_internal_power(TLKFD_voltage, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_intrinsic_fall                  : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_intrinsic(TLKFD_voltage, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_intrinsic_rise                  : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_intrinsic(TLKFD_voltage, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_pin_cap                         : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_pin_cap(TLKFD_voltage, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_recovery_fall                   : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_recovery(TLKFD_voltage, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_recovery_rise                   : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_recovery(TLKFD_voltage, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_removal_fall                    : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_removal(TLKFD_voltage, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_removal_rise                    : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_removal(TLKFD_voltage, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_rise_propagation                : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_propagation(TLKFD_voltage, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_rise_transition                 : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_transition(TLKFD_voltage, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_setup_fall                      : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_setup(TLKFD_voltage, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_setup_rise                      : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_setup(TLKFD_voltage, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_slope_fall                      : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_slope(TLKFD_voltage, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_slope_rise                      : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_slope(TLKFD_voltage, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_nom_process                            : // float( > 0.0 and  < 100.0 ) := 1.0;
      nomOC.process(attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_nom_temperature                        : // float( > -300.0 and  < 500.0 ) := 25.0;
      nomOC.temperature(attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_nom_voltage                            : // float;
      nomOC.voltage(attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_default_operating_conditions           : // string;
      tlib.defaultOpCndName(attr->value()->string_val());
      break;
    case LIBERTY_ATTRENUM_slew_derate_from_library               : // float( > 0.0 and  < 1.0 );
      if (!UParam::ignoreSlewDerateFromLibrary()) {
	tlib.slewDeratingFactor(attr->value()->double_val());
      }
      break;
    case LIBERTY_ATTRENUM_voltage_map                            : // (string, float)
      {
	LibValIter viter = attr->valueIter();
	LibVal* v = viter++;
	const char* voltName = v->string_val();
	v = viter++;
	double voltage = v->double_val();
	tlib.addVoltageMap(voltName, voltage);
      }
      break;
    case LIBERTY_ATTRENUM_bus_naming_style                       : // virtual_attribute;
    case LIBERTY_ATTRENUM_comment                                : // string;
    case LIBERTY_ATTRENUM_current_unit                           : // enum( 100uA, 100mA, 1A, 1uA, 10uA, 1mA, or 10mA );
    case LIBERTY_ATTRENUM_date                                   : // string;
    case LIBERTY_ATTRENUM_default_cell_leakage_power             : // float( > 0.0 );
    case LIBERTY_ATTRENUM_default_connection_class               : // string;
    case LIBERTY_ATTRENUM_default_edge_rate_breakpoint_f0        : // float;
    case LIBERTY_ATTRENUM_default_edge_rate_breakpoint_f1        : // float;
    case LIBERTY_ATTRENUM_default_edge_rate_breakpoint_r0        : // float;
    case LIBERTY_ATTRENUM_default_edge_rate_breakpoint_r1        : // float;
    case LIBERTY_ATTRENUM_default_fall_delay_intercept           : // float := 0.0;
    case LIBERTY_ATTRENUM_default_fall_pin_resistance            : // float( > 0.0 ) := 0.0;
    case LIBERTY_ATTRENUM_default_hold_coefficient               : // float;
    case LIBERTY_ATTRENUM_default_inout_pin_fall_res             : // float( > 0.0 ) := 0.0;
    case LIBERTY_ATTRENUM_default_inout_pin_rise_res             : // float( > 0.0 ) := 0.0;
    case LIBERTY_ATTRENUM_default_intrinsic_fall                 : // float( > 0.0 ) := 1.0;
    case LIBERTY_ATTRENUM_default_intrinsic_rise                 : // float( > 0.0 ) := 1.0;
    case LIBERTY_ATTRENUM_default_leakage_power_density          : // float( > 0.0 );
    case LIBERTY_ATTRENUM_default_max_utilization                : // float( > 0.0 and  < 100.0 );
    case LIBERTY_ATTRENUM_default_min_porosity                   : // virtual_attribute;
    case LIBERTY_ATTRENUM_default_output_pin_fall_res            : // float( > 0.0 ) := 0.0;
    case LIBERTY_ATTRENUM_default_output_pin_rise_res            : // float( > 0.0 ) := 0.0;
    case LIBERTY_ATTRENUM_default_rc_fall_coefficient            : // float;
    case LIBERTY_ATTRENUM_default_rc_rise_coefficient            : // float;
    case LIBERTY_ATTRENUM_default_reference_capacitance          : // float;
    case LIBERTY_ATTRENUM_default_rise_delay_intercept           : // float := 0.0;
    case LIBERTY_ATTRENUM_default_rise_pin_resistance            : // float( > 0.0 ) := 0.0;
    case LIBERTY_ATTRENUM_default_setup_coefficient              : // float;
    case LIBERTY_ATTRENUM_default_slope_fall                     : // float( > 0.0 ) := 0.0;
    case LIBERTY_ATTRENUM_default_slope_rise                     : // float( > 0.0 ) := 0.0;
    case LIBERTY_ATTRENUM_default_wire_load                      : // string;
    case LIBERTY_ATTRENUM_default_wire_load_area                 : // float( > 0.0 );
    case LIBERTY_ATTRENUM_default_wire_load_capacitance          : // float( > 0.0 );
    case LIBERTY_ATTRENUM_default_wire_load_mode                 : // string( "top", "segmented", or "enclosed" );
    case LIBERTY_ATTRENUM_default_wire_load_resistance           : // float( > 0.0 );
    case LIBERTY_ATTRENUM_default_wire_load_selection            : // string;
    case LIBERTY_ATTRENUM_delay_model                            : // virtual_attribute;
    case LIBERTY_ATTRENUM_em_temp_degradation_factor             : // float( > 0.0 );
    case LIBERTY_ATTRENUM_in_place_swap_mode                     : // enum( match_footprint, ignore_footprint, or no_swapping );
    case LIBERTY_ATTRENUM_input_threshold_pct_fall               : // float( > 0.0 and  < 100.0 );
    case LIBERTY_ATTRENUM_input_threshold_pct_rise               : // float( > 0.0 and  < 100.0 );
    case LIBERTY_ATTRENUM_k_process_cell_degradation             : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_process_drive_current                : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_process_fall_delay_intercept         : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_process_fall_pin_resistance          : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_process_min_period                   : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_process_min_pulse_width_high         : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_process_min_pulse_width_low          : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_process_nochange_fall                : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_process_nochange_rise                : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_process_rise_delay_intercept         : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_process_rise_pin_resistance          : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_process_skew_fall                    : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_process_skew_rise                    : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_process_wire_cap                     : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_process_wire_res                     : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_temp_cell_degradation                : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_temp_drive_current                   : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_temp_fall_delay_intercept            : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_temp_fall_pin_resistance             : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_temp_min_period                      : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_temp_min_pulse_width_high            : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_temp_min_pulse_width_low             : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_temp_nochange_fall                   : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_temp_nochange_rise                   : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_temp_rise_delay_intercept            : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_temp_rise_pin_resistance             : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_temp_skew_fall                       : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_temp_skew_rise                       : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_temp_wire_cap                        : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_temp_wire_res                        : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_volt_cell_degradation                : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_volt_drive_current                   : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_volt_fall_delay_intercept            : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_volt_fall_pin_resistance             : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_volt_min_period                      : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_volt_min_pulse_width_high            : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_volt_min_pulse_width_low             : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_volt_nochange_fall                   : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_volt_nochange_rise                   : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_volt_rise_delay_intercept            : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_volt_rise_pin_resistance             : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_volt_skew_fall                       : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_volt_skew_rise                       : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_volt_wire_cap                        : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_volt_wire_res                        : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_key_bit                                : // short( > 1 and  < 20 );
    case LIBERTY_ATTRENUM_key_feature                            : // string;
    case LIBERTY_ATTRENUM_key_file                               : // string;
    case LIBERTY_ATTRENUM_key_seed                               : // int;
    case LIBERTY_ATTRENUM_key_version                            : // float;
    case LIBERTY_ATTRENUM_lsi_pad_fall                           : // float;
    case LIBERTY_ATTRENUM_lsi_pad_rise                           : // float;
    case LIBERTY_ATTRENUM_lsi_rounding_cutoff                    : // float( > 0.0 );
    case LIBERTY_ATTRENUM_lsi_rounding_digit                     : // float( > 0.0 );
    case LIBERTY_ATTRENUM_nom_calc_mode                          : // string;
    case LIBERTY_ATTRENUM_output_threshold_pct_fall              : // float( > 0.0 and  < 100.0 );
    case LIBERTY_ATTRENUM_output_threshold_pct_rise              : // float( > 0.0 and  < 100.0 );
    case LIBERTY_ATTRENUM_piece_type                             : // string( "piece_length", "piece_total_cap", "piece_wire_cap", or "piece_pin_cap" );
    case LIBERTY_ATTRENUM_power_model                            : // virtual_attribute;
    case LIBERTY_ATTRENUM_preferred_input_pad_voltage            : // string;
    case LIBERTY_ATTRENUM_preferred_output_pad_slew_rate_control : // enum( none, high, low, or medium );
    case LIBERTY_ATTRENUM_preferred_output_pad_voltage           : // string;
    case LIBERTY_ATTRENUM_revision                               : // virtual_attribute;
    case LIBERTY_ATTRENUM_simulation                             : // boolean;
    case LIBERTY_ATTRENUM_slew_lower_threshold_pct_fall          : // float( > 0.0 and  < 100.0 );
    case LIBERTY_ATTRENUM_slew_lower_threshold_pct_rise          : // float( > 0.0 and  < 100.0 );
    case LIBERTY_ATTRENUM_slew_upper_threshold_pct_fall          : // float( > 0.0 and  < 100.0 );
    case LIBERTY_ATTRENUM_slew_upper_threshold_pct_rise          : // float( > 0.0 and  < 100.0 );
    case LIBERTY_ATTRENUM_timing_report                          : // boolean;
    case LIBERTY_ATTRENUM_voltage_unit                           : // enum( 1mV, 10mV, 100mV, or 1V );
    case LIBERTY_ATTRENUM_define_cell_area                       : // function( string, string );
    case LIBERTY_ATTRENUM_library_features                       : // function( <unknown_arguments> );
    case LIBERTY_ATTRENUM_piece_define                           : // function( list );
    case LIBERTY_ATTRENUM_routing_layers                         : // function( <unknown_arguments> );
    case LIBERTY_ATTRENUM_technology                             : // function( string );
    default:
      if (warnNoSupport()) {
	printf("Attribute %s (in library %s) on line %d of library file %s"
	       " not supported.\n", attr->name().str(), libName,
	       attr->lineno(), attr->filename());
      }
      break;
    }
  }
}
