// ****************************************************************************
//             Copyright (C) Oasys Design Systems, Inc. 2004 - 2008
//                             All rights reserved.
// ****************************************************************************
#include "LLoad.h"
#include "attr_enum.h"
#include "LLoadFuncParse.h"
#include "LibLoadMsg.h"
#include "TGenOpCnd"
#include "NMVCell"

void
LLoad::processTestPinAttr(const LibGroup* lpin, TLibExternalPinC* tpin,
			  int outCount, LLoadPinMap& pinMap, LLoadBusMap* busses,
			  UAggType* busType, int grpIdx)
{
  LibAttrIter aiter;
  LibAttr* attr;
  Cudd& bddMgr = d_lib.bddMgr();

  aiter = lpin->attrIter();
  while ((attr = aiter++)) {
    switch (attr->typeID()) {
    case LIBERTY_ATTRENUM_function                            : // virtual_attribute;
      {
	TLibScanOutputPinC* outputPin = static_cast<TLibScanOutputPinC*>(tpin);
	outputPin->functionForScanCore(parseFunc(attr->value()->string_val(),
						 bddMgr, pinMap, /*bundles=*/NULL,
						 busses, busType, grpIdx, outCount));
      }
      break;
    case LIBERTY_ATTRENUM_signal_type                         : // string( "test_scan_enable", "test_scan_enable_inverted", "test_scan_clock", "test_scan_clock_a", "test_scan_clock_b", "test_clock", "test_scan_in", "test_scan_in_inverted", "test_scan_out", "test_scan_out_inverted", or "clocked_on_also" );
      {
	TLibExternalPinC* pin = static_cast<TLibExternalPinC*>(tpin);
	uassert (pin->signalType() == TLPST_none);
	UString type(attr->value()->string_val());
	switch (type.keyword()) {
	case K_test_scan_enable:
	  pin->setSignalType(TLPST_test_scan_enable);
	  break;
	case K_test_scan_enable_inverted:
	  pin->setSignalType(TLPST_test_scan_enable_inverted);
	  break;
	case K_test_scan_clock:
	  pin->setSignalType(TLPST_test_scan_clock);
	  break;
	case K_test_scan_clock_a:
	  pin->setSignalType(TLPST_test_scan_clock_a);
	  break;
	case K_test_scan_clock_b:
	  pin->setSignalType(TLPST_test_scan_clock_b);
	  break;
	case K_test_clock:
	  pin->setSignalType(TLPST_test_clock);
	  break;
	case K_test_scan_in:
	  pin->setSignalType(TLPST_test_scan_in);
	  break;
	case K_test_scan_in_inverted:
	  pin->setSignalType(TLPST_test_scan_in_inverted);
	  break;
	case K_test_scan_out:
	  pin->setSignalType(TLPST_test_scan_out);
	  break;
	case K_test_scan_out_inverted:
	  pin->setSignalType(TLPST_test_scan_out_inverted);
	  break;
	case K_clocked_on_also:
	  pin->setSignalType(TLPST_clocked_on_also);
	  break;
	default:
	  uassert(0);
	}
      }
      break;
    case LIBERTY_ATTRENUM_test_output_only                    : // boolean;
      {
	TLibExternalPinC* pin = static_cast<TLibExternalPinC*>(tpin);
	pin->setTestOutOnly(attr->value()->bool_val());
      }
      break;
    }
  }
}

void
LLoad::processPinAttr(const LibGroup* lpin, TLibExternalPinC* tpin, TLibDir dir,
		      LCell& lcell, int pinIdx, int outCount, LLoadPinMap& pinMap,
		      LLoadBundleMap* bundles, LLoadBusMap* busses,
		      UAggType* busType, int grpIdx)
{
  LibAttrIter aiter;
  LibAttr* attr;
  TLib& tlib = *d_lib.tlib();
  Cudd& bddMgr = d_lib.bddMgr();
  TCap cap;

  aiter = lpin->attrIter();
  while ((attr = aiter++)) {
    switch (attr->typeID()) {
    case LIBERTY_ATTRENUM_capacitance                         : // float;
      cap = tlib.convertCap(attr->value()->double_val());
      tpin->setCap(MVCRNR_INITIAL, TEdge::rise, cap);
      tpin->setCap(MVCRNR_INITIAL, TEdge::fall, cap);
      break;
    case LIBERTY_ATTRENUM_clock_gate_clock_pin                : // boolean;
      {
	TLibExternalPinC* cgPin = static_cast<TLibExternalPinC*>(tpin);
	if (attr->value()->bool_val()) {
	  uassert (cgPin->signalType() == TLPST_none);
	  cgPin->setSignalType(TLPST_clock_gate_clock);
	} else {
	  uassert (cgPin->signalType() != TLPST_clock_gate_clock);
	}
      }
      break;
    case LIBERTY_ATTRENUM_clock_gate_enable_pin               : // boolean;
      {
	TLibExternalPinC* cgPin = static_cast<TLibExternalPinC*>(tpin);
	if (attr->value()->bool_val()) {
	  uassert (cgPin->signalType() == TLPST_none);
	  cgPin->setSignalType(TLPST_clock_gate_enable);
	} else {
	  uassert (cgPin->signalType() != TLPST_clock_gate_enable);
	}
      }
      break;
    case LIBERTY_ATTRENUM_clock_gate_obs_pin                  : // boolean;
      {
	TLibExternalPinC* cgPin = static_cast<TLibExternalPinC*>(tpin);
	if (attr->value()->bool_val()) {
	  uassert (cgPin->signalType() == TLPST_none);
	  cgPin->setSignalType(TLPST_clock_gate_obs);
	} else {
	  uassert (cgPin->signalType() != TLPST_clock_gate_obs);
	}
      }
      break;
    case LIBERTY_ATTRENUM_clock_gate_out_pin                  : // boolean;
      {
	TLibExternalPinC* cgPin = static_cast<TLibExternalPinC*>(tpin);
	if (attr->value()->bool_val()) {
	  uassert (cgPin->signalType() == TLPST_none);
	  cgPin->setSignalType(TLPST_clock_gate_out);
	} else {
	  uassert (cgPin->signalType() != TLPST_clock_gate_out);
	}
      }
      break;
    case LIBERTY_ATTRENUM_clock_gate_test_pin                 : // boolean;
      {
	TLibExternalPinC* cgPin = static_cast<TLibExternalPinC*>(tpin);
	if (attr->value()->bool_val()) {
	  uassert (cgPin->signalType() == TLPST_none);
	  cgPin->setSignalType(TLPST_clock_gate_test);
	} else {
	  uassert (cgPin->signalType() != TLPST_clock_gate_test);
	}
      }
      break;
    case LIBERTY_ATTRENUM_fall_capacitance                    : // float;
      cap = tlib.convertCap(attr->value()->double_val());
      tpin->setCap(MVCRNR_INITIAL, TEdge::rise, cap);
      tpin->setCap(MVCRNR_INITIAL, TEdge::fall, cap);
      break;
    case LIBERTY_ATTRENUM_fanout_load                         : // float( > 0.0 ) := default_fanout_load;
      if (dir != TLD_input && dir != TLD_bidir) {
	printf ("ERROR : fanout_load attribute not permitted on non-input "
		"and non-inout pin (file %s, line %d).\n",
		attr->filename(), attr->lineno());
      } else {
	if (dir == TLD_input) {
	  static_cast<TLibInputPinC*>(tpin)->fanoutLoad
	    (attr->value()->double_val());
	} else {
	  static_cast<TLibBidirPinC*>(tpin)->fanoutLoad
	    (attr->value()->double_val());
	}
      }
      break;
    case LIBERTY_ATTRENUM_function                            : // virtual_attribute;
      {
	TLibOutputPinC* outputPin = static_cast<TLibOutputPinC*>(tpin);
	outputPin->function(parseFunc(attr->value()->string_val(),
				      bddMgr, pinMap, bundles, busses, busType,
				      grpIdx, outCount));
      }
      break;
    case LIBERTY_ATTRENUM_is_pad                              : // boolean;
      uassert (attr->value()->bool_val() == true);
      tpin->setIsPad(true);
      break;
#ifdef RT_FPGA
    case LIBERTY_ATTRENUM_clock                              : // boolean;
      if(attr->value()->bool_val() == true) {
	tpin->setIsClock(true);
      }
      break;
#endif
    case LIBERTY_ATTRENUM_level_shifter_enable_pin            : // boolean;
      {
	NCell cell = lcell.ncell();
	NMVCell mvCell = cell.isMVCell();
	uassert(mvCell);
	mvCell.setMvCellType(NMVT_elshifter);
      }
      // no break
    case LIBERTY_ATTRENUM_isolation_cell_enable_pin           : // boolean;
      uassert (attr->value()->bool_val() == true);
      tpin->setIsEnable(true);
      break;
    case LIBERTY_ATTRENUM_max_capacitance                     : // float( > 0.0 ) := default_max_capacitance;
      if (dir != TLD_output && dir != TLD_tristate && dir != TLD_bidir) {
	printf ("ERROR : max_cap attribute not permitted on non-output "
		"and non-inout pin (file %s, line %d).\n",
		attr->filename(), attr->lineno());
      } else {
	static_cast<TLibOutputPinC*>(tpin)->maxCap(tlib.convertCap(attr->value()->double_val()));
      }
      break;
    case LIBERTY_ATTRENUM_max_fanout                          : // float( > 0.0 ) := default_max_fanout;
      if (dir != TLD_output && dir != TLD_tristate && dir != TLD_bidir) {
	printf ("ERROR : max_fanout attribute not permitted on non-output "
		"and non-inout pin (file %s, line %d).\n",
		attr->filename(), attr->lineno());
      } else {
	static_cast<TLibOutputPinC*>(tpin)->maxFanout
	  (attr->value()->double_val());
      }
      break;
    case LIBERTY_ATTRENUM_max_transition                      : // float( > 0.0 ) := default_max_transition;
      {
	double s = attr->value()->double_val();
	TVal maxSlew = tlib.convertTime(s);
	if (maxSlew > TSlew::max()) {
	  LIBLOADMSG(maxSlewTooBig, s);
	  tpin->setMaxSlew(MVCRNR_INITIAL, TSlew::max());
	} else {
	  tpin->setMaxSlew(MVCRNR_INITIAL, maxSlew);
	}
	break;
      }
    case LIBERTY_ATTRENUM_nextstate_type                      : // enum( scan_enable, preset, data, clear, load, or scan_in );
      lcell.setNextStateType(pinIdx, attr->value()->string_val());
      break;
    case LIBERTY_ATTRENUM_rise_capacitance                    : // float;
      cap = tlib.convertCap(attr->value()->double_val());
      tpin->setCap(MVCRNR_INITIAL, TEdge::rise, cap);
      tpin->setCap(MVCRNR_INITIAL, TEdge::fall, cap);
      break;
    case LIBERTY_ATTRENUM_signal_type                         : // string( "test_scan_enable", "test_scan_enable_inverted", "test_scan_clock", "test_scan_clock_a", "test_scan_clock_b", "test_clock", "test_scan_in", "test_scan_in_inverted", "test_scan_out", "test_scan_out_inverted", or "clocked_on_also" );
      {
	TLibExternalPinC* pin = static_cast<TLibExternalPinC*>(tpin);
	uassert (pin->signalType() == TLPST_none);
	UString type(attr->value()->string_val());
	switch (type.keyword()) {
	case K_test_scan_enable:
	  pin->setSignalType(TLPST_test_scan_enable);
	  break;
	case K_test_scan_enable_inverted:
	  pin->setSignalType(TLPST_test_scan_enable_inverted);
	  break;
	case K_test_scan_clock:
	  pin->setSignalType(TLPST_test_scan_clock);
	  break;
	case K_test_scan_clock_a:
	  pin->setSignalType(TLPST_test_scan_clock_a);
	  break;
	case K_test_scan_clock_b:
	  pin->setSignalType(TLPST_test_scan_clock_b);
	  break;
	case K_test_clock:
	  pin->setSignalType(TLPST_test_clock);
	  break;
	case K_test_scan_in:
	  pin->setSignalType(TLPST_test_scan_in);
	  break;
	case K_test_scan_in_inverted:
	  pin->setSignalType(TLPST_test_scan_in_inverted);
	  break;
	case K_test_scan_out:
	  pin->setSignalType(TLPST_test_scan_out);
	  break;
	case K_test_scan_out_inverted:
	  pin->setSignalType(TLPST_test_scan_out_inverted);
	  break;
	case K_clocked_on_also:
	  pin->setSignalType(TLPST_clocked_on_also);
	  break;
	default:
	  uassert(0);
	}
      }
      break;
    case LIBERTY_ATTRENUM_test_output_only                    : // boolean;
      {
	TLibExternalPinC* pin = static_cast<TLibExternalPinC*>(tpin);
	pin->setTestOutOnly(attr->value()->bool_val());
	break;
      }
    case LIBERTY_ATTRENUM_three_state                         : // virtual_attribute;
      {
	TLibTristatePinC* tristatePin = static_cast<TLibTristatePinC*>(tpin);
	tristatePin->tristate(parseFunc(attr->value()->string_val(),
					bddMgr, pinMap, bundles, busses, busType,
					grpIdx, outCount));
      }
      break;
    case LIBERTY_ATTRENUM_input_signal_level                  : // string;
    case LIBERTY_ATTRENUM_output_signal_level                 : // string;
      {
	const char* s = attr->value()->string_val();
	TVal volt     = tlib.voltageMap(s);
	tpin->setSignalLevel(MVCRNR_INITIAL, volt);

	// here we try to guess the related_power
	int relatedPower = -1;
	bool unique = true;
	for (UAggIter iter = lcell.allPGPins().nocheck().msbToLsb(); iter; iter++) {
	  if (lcell.pgpin(iter)->voltageName() == s) {
	    if (relatedPower == -1) {
	      relatedPower = iter.current();
	    } else {
	      unique = false;
	    }
	  }
	}
	if (relatedPower >= 0 && unique) {
	  lcell.setRelatedPower(pinIdx, relatedPower,
				lcell.totalPinCount(), lcell.totalPGPinCount());
	}
      }
      break;
    case LIBERTY_ATTRENUM_related_power_pin                   : // string;
    case LIBERTY_ATTRENUM_related_ground_pin                  : // string;
      {
	const char* s = attr->value()->string_val();

	// here we find the PG pin
	int relatedPG = -1;
	TLibPGPin* pgPin = NULL;
	for (UAggIter iter = lcell.allPGPins().nocheck().msbToLsb(); iter; iter++) {
	  if (lcell.pgpin(iter)->name() == s) {
	    pgPin = lcell.pgpin(iter);
	    relatedPG = iter.current();
	    break;
	  }
	}

	if (relatedPG >= 0) {
	  if (attr->typeID() == LIBERTY_ATTRENUM_related_power_pin) {
	    lcell.setRelatedPower(pinIdx, relatedPG,
				  lcell.totalPinCount(), lcell.totalPGPinCount());
	    if (tpin->signalLevel(MVCRNR_INITIAL).isFuzzyEqual(UFloat(0))) {
	      // set the signal level only if it has not been specified
	      UString vName = pgPin->voltageName();
	      TVal volt     = tlib.voltageMap(vName);
	      tpin->setSignalLevel(MVCRNR_INITIAL, volt);
	    }
	  } else {
	    lcell.setRelatedGround(pinIdx, relatedPG,
				   lcell.totalPinCount(), lcell.totalPGPinCount());
	  }
	}
      }
      break;
    case LIBERTY_ATTRENUM_retention_pin                       : // (string, string); 
    case LIBERTY_ATTRENUM_power_gating_pin                    : // (string, int);
      {
	LibValIter viter = attr->valueIter();
	LibVal* v = viter++;
	UString pinClassStr = v->string_val();
	TRetPinClass pinClass = TRETPC_save;
	bool ok = true;

	// check pinClass
	switch (pinClassStr.keyword()) {
	case K_restore:
	case K_power_pin_1:
	  pinClass = TRETPC_restore;
	  break;
	case K_save:
	case K_power_pin_2:
	  pinClass = TRETPC_save;
	  break;
	case K_save_restore:
	case K_power_pin_3:
	  pinClass = TRETPC_save_restore;
	  break;
	default:
	  LIBLOADMSG(invalidRetentionPinClass, lcell.pinName(pinIdx).str(),
		     lcell.ncell()._name().str(), pinClassStr.str());
	  ok = false;
	}

	// check disableValue
	v = viter++;
	int disValue = -1;
	if (attr->typeID() == LIBERTY_ATTRENUM_retention_pin) {
	  UString dis      = v->string_val();
	  disValue = atoi(dis);
	} else {
	  disValue = v->int_val();
	}
	if (disValue != 1 && disValue != 0) {
	  LIBLOADMSG(invalidDisableValue, lcell.pinName(pinIdx).str(),
		     lcell.ncell()._name().str(), disValue);
	  ok = false;
	}

	if (ok) {
	  lcell.setRetentionPin(pinClass, pinIdx, disValue);
	}
      }
      break;
#ifndef RT_FPGA
    case LIBERTY_ATTRENUM_clock                               : // boolean;
#endif
    case LIBERTY_ATTRENUM_complementary_pin                   : // string;
    case LIBERTY_ATTRENUM_connection_class                    : // string;
    case LIBERTY_ATTRENUM_dcm_timing                          : // boolean;
    case LIBERTY_ATTRENUM_direction                           : // enum( inout, input, internal, output, or tristate );
    case LIBERTY_ATTRENUM_dont_fault                          : // string( "sa0", "sa1", or "sa01" );
    case LIBERTY_ATTRENUM_drive_current                       : // float( > 0.0 );
    case LIBERTY_ATTRENUM_driver_type                         : // virtual_attribute;
    case LIBERTY_ATTRENUM_edge_rate_breakpoint_f0             : // float;
    case LIBERTY_ATTRENUM_edge_rate_breakpoint_f1             : // float;
    case LIBERTY_ATTRENUM_edge_rate_breakpoint_r0             : // float;
    case LIBERTY_ATTRENUM_edge_rate_breakpoint_r1             : // float;
    case LIBERTY_ATTRENUM_edge_rate_fall                      : // float;
    case LIBERTY_ATTRENUM_edge_rate_load_fall                 : // float;
    case LIBERTY_ATTRENUM_edge_rate_load_rise                 : // float;
    case LIBERTY_ATTRENUM_edge_rate_rise                      : // float;
    case LIBERTY_ATTRENUM_fall_current_slope_after_threshold  : // float;
    case LIBERTY_ATTRENUM_fall_current_slope_before_threshold : // float;
    case LIBERTY_ATTRENUM_fall_time_after_threshold           : // float( > 0.0 );
    case LIBERTY_ATTRENUM_fall_time_before_threshold          : // float( > 0.0 );
    case LIBERTY_ATTRENUM_fault_model                         : // string( "XX", "11", "00", "X1", "X0", "1X", "10", "0X", or "01" );
    case LIBERTY_ATTRENUM_fsim_map                            : // string;
    case LIBERTY_ATTRENUM_has_builtin_pad                     : // boolean;
    case LIBERTY_ATTRENUM_hysteresis                          : // boolean;
    case LIBERTY_ATTRENUM_input_map                           : // virtual_attribute;
    case LIBERTY_ATTRENUM_input_voltage                       : // string;
    case LIBERTY_ATTRENUM_internal_node                       : // string;
    case LIBERTY_ATTRENUM_inverted_output                     : // boolean;
    case LIBERTY_ATTRENUM_lsi_pad                             : // boolean;
    case LIBERTY_ATTRENUM_max_time_borrow                     : // float( > 0.0 );
    case LIBERTY_ATTRENUM_min_capacitance                     : // float( > 0.0 );
    case LIBERTY_ATTRENUM_min_fanout                          : // float( > 0.0 );
    case LIBERTY_ATTRENUM_min_period                          : // float;
    case LIBERTY_ATTRENUM_min_pulse_width_high                : // float( > 0.0 );
    case LIBERTY_ATTRENUM_min_pulse_width_low                 : // float( > 0.0 );
    case LIBERTY_ATTRENUM_min_transition                      : // float( > 0.0 );
    case LIBERTY_ATTRENUM_multicell_pad_pin                   : // boolean;
    case LIBERTY_ATTRENUM_output_voltage                      : // string;
    case LIBERTY_ATTRENUM_pin_func_type                       : // enum( active_rising, active_falling, clock_enable, active_high, or active_low );
    case LIBERTY_ATTRENUM_prefer_tied                         : // string;
    case LIBERTY_ATTRENUM_primary_output                      : // boolean;
    case LIBERTY_ATTRENUM_pulling_current                     : // float;
    case LIBERTY_ATTRENUM_pulling_resistance                  : // float( > 0.0 );
    case LIBERTY_ATTRENUM_reference_capacitance               : // float;
    case LIBERTY_ATTRENUM_rise_current_slope_after_threshold  : // float;
    case LIBERTY_ATTRENUM_rise_current_slope_before_threshold : // float;
    case LIBERTY_ATTRENUM_rise_time_after_threshold           : // float( > 0.0 );
    case LIBERTY_ATTRENUM_rise_time_before_threshold          : // float( > 0.0 );
    case LIBERTY_ATTRENUM_slew_control                        : // enum( none, high, low, or medium );
    case LIBERTY_ATTRENUM_state_function                      : // virtual_attribute;
    case LIBERTY_ATTRENUM_vhdl_name                           : // string;
    case LIBERTY_ATTRENUM_x_function                          : // virtual_attribute;
    default:
      if (warnNoSupport()) {
	printf("Attribute %s (in pin group) on line %d of library file %s"
	       " not supported.\n", attr->name().str(), attr->lineno(),
	       attr->filename());
      }
      break;
    };
  }
}

void
LLoad::processPGPinAttr(const LibGroup* lpin, TLibPGPinC* tpin,
			TLibDir dir,
			LCell& lcell, int pinIdx)
{
  LibAttrIter aiter;
  LibAttr* attr;
  TLib& tlib = *d_lib.tlib();
  UFloat volt;
  UString pgType;

  aiter = lpin->attrIter();
  while ((attr = aiter++)) {
    switch (attr->typeID()) {
    case LIBERTY_ATTRENUM_voltage_name:                         // string;
      volt = tlib.voltageMap(attr->value()->string_val());
      tpin->setVoltage(MVCRNR_INITIAL, volt);
      tpin->setVoltageName(attr->value()->string_val());
      break;
    case LIBERTY_ATTRENUM_pg_type:                              // string
      pgType = attr->value()->string_val();
      tpin->setPGType(pgType);
      break;
    case LIBERTY_ATTRENUM_direction:                            // string
      // already processed
      break;
    case LIBERTY_ATTRENUM_std_cell_main_rail:                   // boolean
      {
	bool isMainR = attr->value()->bool_val();
	if (isMainR) {
	  tpin->setMainRail(isMainR);
	}
	break;
      }
    case LIBERTY_ATTRENUM_switch_function:                      // string 
    case LIBERTY_ATTRENUM_pg_function:                          // string
    default:
      if (warnNoSupport()) {
	printf("Attribute %s (in pin group) on line %d of library file %s"
	       " not supported.\n", attr->name().str(), attr->lineno(),
	       attr->filename());
      }
      break;
    };
  }
}
