#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include "mymalloc.h"
#include "LibVal.h"
#include "LibLoadMsg.h"
#ifdef _WIN32
#include "oasys_port.h"
#endif

LibVal::~LibVal() {
  if (d_type == LIBERTY__VAL_STRING) {
    my_free (d_u.string_val);
  }
}

const char*
LibVal::typeStr() const
{
  switch (d_type) {
  case LIBERTY__VAL_UNDEFINED: return "undefined";
  case LIBERTY__VAL_INT      : return "integer";
  case LIBERTY__VAL_DOUBLE   : return "double";
  case LIBERTY__VAL_BOOLEAN  : return "boolean";
  case LIBERTY__VAL_STRING   : return "string";
  default                    : return "";
  }
  return "";
}

const char*
LibVal::string_val() const
{
  if (d_type == LIBERTY__VAL_STRING) {
    return d_u.string_val;
  } else {
    LIBLOADMSG(unexpectedLibValueType, "string", typeStr());
    return "";
  }
}

int
LibVal::int_val() const
{
  if (d_type == LIBERTY__VAL_DOUBLE) {
    return (int)d_u.double_val;
  } else if (d_type == LIBERTY__VAL_INT) {
    return d_u.int_val;
  } else if (d_type == LIBERTY__VAL_STRING) {
    errno = 0;
    char *end;
    long val = strtol(d_u.string_val, &end, 10);
    if (!errno && *end == '\0' &&
	val <= (long)INT_MAX && val >= (long)INT_MIN) {
      return (int)val;
    }
  }
  LIBLOADMSG(unexpectedLibValueType, "integer", typeStr());
  return -1;
}

double
LibVal::double_val() const
{
  if (d_type == LIBERTY__VAL_DOUBLE) {
    return d_u.double_val;
  } else if (d_type == LIBERTY__VAL_INT) {
    return (double)d_u.int_val;
  } else if (d_type == LIBERTY__VAL_STRING) {
    errno = 0;
    char *end;
    double val = strtod(d_u.string_val, &end);
    if (!errno && *end == '\0') {
      return val;
    }
  }
  LIBLOADMSG(unexpectedLibValueType, "double", typeStr());
  return -1;
}

bool
LibVal::bool_val() const
{
  if (d_type == LIBERTY__VAL_BOOLEAN) {
    return d_u.bool_val;
  } else if (d_type == LIBERTY__VAL_STRING) {
    if (!strcasecmp(d_u.string_val, "true")) {
      return true;
    } else if (!strcasecmp(d_u.string_val, "false")) {
      return false;
    }
  }
  LIBLOADMSG(unexpectedLibValueType, "boolean", typeStr());
  return d_u.bool_val;
}

void
LibVal::dump(FILE* f, const char* prepend, const char* append)
{
  switch (d_type) {
  case LIBERTY__VAL_STRING:
    fprintf(f, "%s\"%s\"%s", prepend, string_val(), append);
    break;
  case LIBERTY__VAL_DOUBLE:
    fprintf(f, "%s%g%s", prepend, double_val(), append);
    break;
  case LIBERTY__VAL_INT:
    fprintf(f, "%s%d%s", prepend, int_val(), append);
    break;
  case LIBERTY__VAL_BOOLEAN:
    fprintf(f, "%s%s%s", prepend, (bool_val() ? "true" : "false"), append);
    break;
  default:
    break;
  }
}

