 /******************************************************************************
    Copyright (c) 1996-2000 Synopsys, Inc.    ALL RIGHTS RESERVED

  The contents of this file are subject to the restrictions and limitations
  set forth in the SYNOPSYS Open Source License Version 1.0  (the "License"); 
  you may not use this file except in compliance with such restrictions 
  and limitations. You may obtain instructions on how to receive a copy of 
  the License at

  http://www.synopsys.com/partners/tapin/tapinprogram.html. 

  Software distributed by Original Contributor under the License is 
  distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either 
  expressed or implied. See the License for the specific language governing 
  rights and limitations under the License.

******************************************************************************/

#include <stdio.h>
#include <sys/types.h>
#include <time.h>
#include <string.h>
#include <stdlib.h>

#include "UFile"
#include "LLoad.h"
#include "TLibDelay"
#include "TGenOpCnd"
#include "LLoadFuncParse.h"
#include "NDb"
#include "NTargetLib"
#include "NCell"
#include "NRef"
#include "NRefFamily"
#include "NStub"
#include "UParam"

namespace UParam {
  extern Variable<bool> ignoreLogicalOnlyCellsForSynthesis;
}

inline static
bool operator==(const UFloat& lhs, const UFloat& rhs)
{
  return((pint_t)lhs == (pint_t)rhs);
}

static void print_version(void)
{
  printf("liberty_parse version %s\n\n", VERSION);
}

static void usage(void)
{
  print_version();
	
  printf("Usage: liberty_parse [options] filename\n\n");
  printf("Where [options are:\n\n");
  printf("--version  print version info and quit\n");
  printf("--help     print usage and quit\n");
  printf("\n");
}

void
t3()
{
  /* a simple test case */
  LibErrorT e;
  LibGroup* g1, *g2;
  LibGroup* masterGroup;
  LibNameIter niter;
  LibFileInfo f(0);
	
  LibParse::init(UControl());
  masterGroup = LibParse::group();
  g1 = masterGroup->createGroup("lib1", "library", f, e);
  g2 = g1->createGroup("cell1", "cell", f, e);
	
  g1->addName("lib2",e);
  g1->addName("lib3",e);
  g1->addName("lib4",e);
  g1->addName("lib5",e);
	
  printf("Before\n");
  niter = g1->nameIter();
  while( niter )
    {
      printf("Lib Name: %s\n", (str_t)*niter++);
    }
  niter = g2->nameIter();
  while( niter) 
    {
      printf("Cell Name: %s\n",(str_t)*niter++);
    }
  printf("======================================\n\n");
  
  g1->deleteName("lib5",e);
  g1->addName("lib5a",e);
	
  g2->deleteName("cell1",e);
  g2->addName("cell1a",e);
	
  printf("After\n");
  niter = g1->nameIter();
  while( niter )
    {
      printf("Lib Name: %s\n", (str_t)*niter++);
    }
  niter = g2->nameIter();
  while( niter )
    {
      printf("Cell Name: %s\n",(str_t)*niter++);
    }
  printf("======================================\n\n");
  
  LibParse::quit();
}  

void
move()
{
  LibErrorT err;
  LibGroup* lib, *g1, *g2, *g3;
  LibGroup* masterGroup;
  LibAttr* attr1;
  LibFileInfo f;
	
  LibParse::init(UControl());

  masterGroup = LibParse::group();

  lib = masterGroup->createGroup("lib1", "lib", f, err);

  attr1 = lib->createAttr("default_operating_conditions", LIBERTY__SIMPLE, f, err);
  attr1->value(new LibVal(LIBERTY__VAL_STRING));
  attr1->value()->string_val(uStrdup("oc1"));
	
  g1 = lib->createGroup("g1", "cell", f, err);
  g2 = lib->createGroup("g2", "cell", f, err);
  g3 = lib->createGroup("g3", "cell", f, err);
	
  g3->moveAfter(g1);
	
  LibParse::write(UFile::tmpDir() << "g1g3g2.lib", lib);
	
  g2->moveBefore(g1);
	
  LibParse::write(UFile::tmpDir() << "g2g1g3.lib", lib);

  LibParse::quit();
}

/* test program for liberty parser */

void readLiberty(char* filename, bool warnNoSupport, bool verbose)
{
  bool write_out = 0, print_stats = 1;

  NDb db(filename);
  LLoad libLoader(filename, db, write_out, print_stats, warnNoSupport);
  if (verbose) {
      db.targetLib().add(libLoader.library());
      db.targetLib().dump();
  }
  UMem::printStats();
  db.destroy();
}

void
testExample(char* filename)
{
  int write_out = 0, print_stats = 1;

  NDb db(filename);
  LLoad libLoader(filename, db, write_out, print_stats, false);

  // check normalization of lookup table template indices
  // and timing tables;

  NLib lib = libLoader.library();
  if (!lib) {
    return;
  }
  NTargetLib target = db.targetLib();
  target.add(lib);
  target.dump(/*familiesOnly=*/false);
  NCell an2a = lib.findCell("AN2A");
  TLibExternalPinC* tpin = an2a.pin(0)->external();

  TLibDelay& d = *tpin->delays(1, MVCRNR_COMMON, TLibCorner::maximum)->begin().object();
  assert (d.delayType(TEdge::rise) == TLDT_cellNTrans);
  TLibDelayTable* rtable = d.table(TEdge::rise);
  
  TLutIndexTable *idxTable = rtable->idxTable();
  assert (idxTable->varCount() == 2);
  assert (idxTable->varPos(0) == 1);
  assert (idxTable->varPos(1) == 0);
  assert (idxTable->varType(0) == TLVT_total_output_net_capacitance);
  assert (idxTable->varType(1) == TLVT_input_net_transition);

  TLib& tlib = *lib.tlib();
  double mult[3];
  mult[0] = tlib.capUnitConversion();
  mult[1] = tlib.timeUnitConversion();
  for (uint i = 0; i < idxTable->varCount(); i++) {
    TLutIndices& index = idxTable->indices(i);
    for (uint j = 0; j < index.count(); j++) {
      assert (index[j] == TVal((2 * j + i + 1) * mult[i]));
    }
  }

  {
    TLutIndices& index0 = idxTable->indices(0);
    TLutIndices& index1 = idxTable->indices(1);
    for (uint i = 0; i < index0.count(); i++) {
      for (uint j = 0; j < index1.count(); j++) {
	assert (rtable->value(index0[i], index1[j]) ==
		 TVal((i + j * index0.count() + 1) * tlib.timeUnitConversion()));
      }
    }
  }


  assert (d.delayType(TEdge::fall) == TLDT_cellNTrans);
  TLibDelayTable* ftable = d.table(TEdge::fall);
  
  idxTable = ftable->idxTable();
  assert (idxTable->varCount() == 3);
  assert (idxTable->varPos(0) == 2);
  assert (idxTable->varPos(1) == 0);
  assert (idxTable->varPos(2) == 1);
  assert (idxTable->varType(0) == TLVT_total_output_net_capacitance);
  assert (idxTable->varType(1) == TLVT_input_net_transition);
  assert (idxTable->varType(2) == TLVT_related_out_total_output_net_capacitance);

  mult[2] = tlib.capUnitConversion();
  for (uint i = 0; i < idxTable->varCount(); i++) {
    TLutIndices& index = idxTable->indices(i);
    for (uint j = 0; j < index.count(); j++) {
      assert (index[j] == TVal((2 * j + i + 1) * mult[i]));
    }
  }

  {
    TLutIndices& index0 = idxTable->indices(0);
    TLutIndices& index1 = idxTable->indices(1);
    TLutIndices& index2 = idxTable->indices(2);

    for (uint i = 0; i < index0.count(); i++) {
      for (uint j = 0; j < index1.count(); j++) {
	for (uint k = 0; k < index2.count(); k++) {
	  assert (ftable->value(index0[i], index1[j], index2[k]) ==
		   TVal((j + 1 + (i + k * index0.count()) * index1.count()) * tlib.timeUnitConversion()));
	}
      }
    }
  }

  // test cell family classification
  NCell an2 = lib.findCell("AN2");
  NCell or2 = lib.findCell("OR2");
  NCell fa1 = lib.findCell("FA1");
  NCell fa2 = lib.findCell("FA2");
  NCell mx1 = lib.findCell("MX1");
  NCell mx2 = lib.findCell("MX2");
  NCell mx3 = lib.findCell("MX3");
  NCell mx4 = lib.findCell("MX4");
  NCell mx5 = lib.findCell("MX5");
  NCell mx6 = lib.findCell("MX6");
  NCell mx7 = lib.findCell("MX7");

  assert (an2.isFamily(an2a, target));
  assert (fa1.isFamily(fa2, target));
  assert (mx3.isFamily(mx4, target));
  assert (mx1.isFamily(mx2, target));
  assert (!fa1.isFamily(an2, target));
  assert (!mx3.isFamily(fa2, target));
  assert (!an2a.isFamily(mx4, target));

  // even though or2 has no equivalent cell in the library,
  // since all its output pin(s) has a function defined, it should be
  // in the cellFamilies hash table
  assert (or2.getFamily(target));
  assert (target.findRefFamily(or2));

  // mx5 and mx6 have one undefined output each and should not be even
  // in the cellFamilies hashtable
   assert (!mx5.getFamily(target));
   assert (!mx6.getFamily(target));
   assert (!target.findRefFamily(mx5));
   assert (!target.findRefFamily(mx6));

  // mx7 has a different user_function_class defined so it should
  // not have any equivalent cell, but should still be in the hash
  // table
  assert (!mx7.getFamily(target).references().moreThan(1));
  assert (!target.findRefFamily(mx7).references().moreThan(1));

  // test the search for a specific cell family that implements a
  // given set of functions
  // need to enclose the following in a separate body so the BDD
  // references by the temporary variables will be properly dereferenced
  // before the database and Cudd mgr is destroyed
  {
    Cudd& mgr = db.bddMgr();
    BDD a = mgr.bddVar(0);
    BDD b = mgr.bddVar(1);
    BDD c = mgr.bddVar(2);
  
    BDDvector f(2), t(2);

    f[0] = b & a;
    assert (target.findRefFamily(f, t, 1, 0, 2) ==
	     target.findRefFamily(an2));

    // need to be in the right order; need to sort the
    // set of functions based on BDD pointers
    f[1] = a ^ b ^ c;
    f[0] = a * b + a * c + b * c;
    Cudd::qSortFunctionVectors(f, t, 0, 1);
    assert (target.findRefFamily(f, t, 2, 0, 3) ==
	     target.findRefFamily(fa1));
    
  }

  db.destroy();
  printf("Done....\n");
}

typedef UHashMap<UString, int> Nm2IdxMap;

void
testFuncParse()
{
  Cudd mgr;
  Nm2IdxMap idxMap;
  const char* str;

  idxMap["Z"] = 0;
  idxMap["ZN"] = 1;
  idxMap["A"] = 2;
  idxMap["A1"] = 3;
  idxMap["B"] = 4;
  idxMap["C"] = 5;
  idxMap["CI0NV"] = 6;
  idxMap["D"] = 7;

  char* inames[] = {(char*)"A", (char*)"A1", (char*)"B", (char*)"C", (char*)"CI0NV", (char*)"D"};
  char* onames[] = {(char*)"Z", (char*)"Z1"};

  BDDvector x(1, &mgr);
  BDD& f = x[0];
  BDD tmp;
  BDD A = mgr.bddVar(0);
  BDD A1 = mgr.bddVar(1);
  BDD B = mgr.bddVar(2);
  BDD C = mgr.bddVar(3);
  BDD CI0NV = mgr.bddVar(4);
  BDD D = mgr.bddVar(5);

  // positive tests

  str = "(A ((A1 D) | ((!A1) CI0NV)) | (!A) ((A1 C) | ((!A1) B)))";
  printf ("\nf = %s : \n\n", str);
  f = parseFunc(str, mgr, idxMap, NULL, NULL, NULL, -1, 2 /*idxOffset*/);
  assert (f.ptr());
  tmp = (A *((A1 * D) + ((~A1) * CI0NV)) + ((~A) * ((A1 * C) + ((~A1) * B))));
  assert(f == tmp);
  tmp = (A * A1 * D) + (A * (~A1) * CI0NV) + ((~A) * A1 * C) +  ((~A) * (~A1) * B);
  assert(f == tmp);
  x.DumpBlif(inames, onames);
  printf("\n");
  f.PrintMinterm();
  
  str = "(A | B)' (C | D)";
  printf ("\nf = %s : \n\n", str);
  f = parseFunc(str, mgr, idxMap, NULL, NULL, NULL, -1, 2 /*idxOffset*/);
  assert (f.ptr());
  tmp = ~(A + B) * (C + D);
  assert(f == tmp);
  tmp = ~A * ~B * C + ~A * ~B * D;
  assert(f == tmp);
  x.DumpBlif(inames, onames);
  printf("\n");
  f.PrintMinterm();
  
  str = "(((A ^ B) !CI0NV) | (A B))";
  printf ("\nf = %s : \n\n", str);
  f = parseFunc(str, mgr, idxMap, NULL, NULL, NULL, -1, 2 /*idxOffset*/);
  assert (f.ptr());
  tmp = ((A ^ B) * ~CI0NV) + (A * B);
  assert(f == tmp);
  x.DumpBlif(inames, onames);
  printf("\n");
  f.PrintMinterm();
  
  str = "A(B'+C )";
  printf ("\nf = %s : \n\n", str);
  f = parseFunc(str, mgr, idxMap, NULL, NULL, NULL, -1, 2 /*idxOffset*/);
  assert (f.ptr());
  tmp = A * (~B + C);
  assert(f == tmp);
  x.DumpBlif(inames, onames);
  printf("\n");
  f.PrintMinterm();
  
  str = "A A1 B' C ";
  printf ("\nf = %s : \n\n", str);
  f = parseFunc(str, mgr, idxMap, NULL, NULL, NULL, -1, 2 /*idxOffset*/);
  assert (f.ptr());
  tmp = A * A1 * ~B * C;
  assert(tmp == f);
  x.DumpBlif(inames, onames);
  printf("\n");
  f.PrintMinterm();
  
  str = "A + B + A1' + C ";
  printf ("\nf = %s : \n\n", str);
  f = parseFunc(str, mgr, idxMap, NULL, NULL, NULL, -1, 2 /*idxOffset*/);
  assert (f.ptr());
  tmp = A + B + ~A1 + C;
  assert(tmp == f);
  x.DumpBlif(inames, onames);
  printf("\n");
  f.PrintMinterm();
  
  str = "A1' ^ !B | A C";
  printf ("\nf = %s : \n\n", str);
  f = parseFunc(str, mgr, idxMap, NULL, NULL, NULL, -1, 2);
  assert (f.ptr());
  tmp = (~A1 ^ ~B) + (A & C);
  assert(tmp == f);
  x.DumpBlif(inames, onames);
  printf("\n");
  f.PrintMinterm();

  str = "(A & (B | C))'";
  printf ("\nf = %s : \n\n", str);
  f = parseFunc(str, mgr, idxMap, NULL, NULL, NULL, -1, 2);
  assert (f.ptr());
  tmp = ~(A * (B + C));
  assert(tmp == f);
  x.DumpBlif(inames, onames);
  printf("\n");
  f.PrintMinterm();

  str = "A ^ (B | C)' & A1";
  printf ("\nf = %s : \n\n", str);
  f = parseFunc(str, mgr, idxMap, NULL, NULL, NULL, -1, 2);
  assert (f.ptr());
  tmp = (A ^ ~(B + C)) * A1;
  assert(tmp == f);
  x.DumpBlif(inames, onames);
  printf("\n");
  f.PrintMinterm();

  str = "(((A & B) | B' &(A1 ^ C)))";
  printf ("\nf = %s : \n\n", str);
  f = parseFunc(str, mgr, idxMap, NULL, NULL, NULL, -1, 2);
  assert (f.ptr());
  tmp = (A * B) + (~B * (A1 ^ C));
  assert(f == tmp);
  x.DumpBlif(inames, onames);
  printf("\n");
  f.PrintMinterm();

  str = "(((A & B) | B' &!(A1 ^ C)))";
  printf ("\nf = %s : \n\n", str);
  f = parseFunc(str, mgr, idxMap, NULL, NULL, NULL, -1, 2);
  assert (f.ptr());
  tmp = (A * B) + (~B * ~(A1 ^ C));
  assert(f == tmp);
  x.DumpBlif(inames, onames);
  printf("\n");
  f.PrintMinterm();

  str = "(((A & B!A1 ^ C)))";
  printf ("\nf = %s : \n\n", str);
  f = parseFunc(str, mgr, idxMap, NULL, NULL, NULL, -1, 2);
  assert (f.ptr());
  tmp = (A * B) * (~A1 ^ C);
  assert(f == tmp);
  x.DumpBlif(inames, onames);
  printf("\n");
  f.PrintMinterm();

  str = "A B + B' A1 + A ^ C &A1' + A'&C";
  printf ("\nf = %s : \n\n", str);
  f = parseFunc(str, mgr, idxMap, NULL, NULL, NULL, -1, 2);
  assert (f.ptr());
  tmp = (A * B) + (~B * A1) + ((A ^ C) * ~A1) + (~A * C);
  assert(f == tmp);
  x.DumpBlif(inames, onames);
  printf("\n");
  f.PrintMinterm();

  str = "(A ^ B ^ A1)!C";
  printf ("\nf = %s : \n\n", str);
  f = parseFunc(str, mgr, idxMap, NULL, NULL, NULL, -1, 2);
  assert (f.ptr());
  tmp = ((A ^ B) ^ A1) * ~C;
  assert(f == tmp);
  x.DumpBlif(inames, onames);
  printf("\n");
  f.PrintMinterm();

  str = "0";
  printf ("\nf = %s : \n\n", str);
  f = parseFunc(str, mgr, idxMap, NULL, NULL, NULL, -1, 2);
  assert (f.ptr());
  assert (f == mgr.bddZero());
  x.DumpBlif(inames, onames);
  printf("\n");
  f.PrintMinterm();
  
  str = "1";
  printf ("\nf = %s : \n\n", str);
  f = parseFunc(str, mgr, idxMap, NULL, NULL, NULL, -1, 2);
  assert (f.ptr());
  assert (f == mgr.bddOne());
  x.DumpBlif(inames, onames);
  printf("\n");
  f.PrintMinterm();

  // negative tests

  str = "(!(A & C)";
  printf ("\nf = %s : \n\n", str);
  f = parseFunc(str, mgr, idxMap, NULL, NULL, NULL, -1, 2);
  assert (!f.ptr());

  str = "(B | C ))";
  printf ("\nf = %s : \n\n", str);
  f = parseFunc(str, mgr, idxMap, NULL, NULL, NULL, -1, 2);
  assert (!f.ptr());

  str = "A & C | B1'";
  printf ("\nf = %s : \n\n", str);
  f = parseFunc(str, mgr, idxMap, NULL, NULL, NULL, -1, 2);
  assert (!f.ptr());

  str = "A & B | 0";
  printf ("\nf = %s : \n\n", str);
  f = parseFunc(str, mgr, idxMap, NULL, NULL, NULL, -1, 2);
  assert (!f.ptr());

  str = "(A|B)1";
  printf ("\nf = %s : \n\n", str);
  f = parseFunc(str, mgr, idxMap, NULL, NULL, NULL, -1, 2);
  assert (!f.ptr());

  str = "A B | A' C";
  f = parseFunc(str, mgr, idxMap, NULL, NULL, NULL, -1, 2);
  assert (f.ptr());
  assert (!(fabs(f.CountMinterm(3) - 4) > 0));
  assert (f.BooleanDiff(0) == (B ^ C));
  assert (f.BooleanDiff(2) == A);
  assert (f.BooleanDiff(3) == ~A);
  assert (f.Constrain(A) == B);
  assert (f.Constrain(~A) == C);
  assert (f.Constrain(B) == (A + C));
  assert (f.Constrain(~C) == (A * B));
}
  
void
runtests()
{
  char filename[1000];
  char *sbroot = getenv("SBROOT");

  t3();
  move();
  filename[0]='\0';
  if (sbroot) {
    strcat(filename, sbroot);
    strcat(filename, "/src/");
  }
  strcat(filename, "../regress/libraries/example.lib");
  testExample(filename);
  testFuncParse();
}

int main(int argc, char **argv)
{
  int version = 0;
  int help = 0;
  char filename[1000];
  int i;
  int mempool_quiet = 1;
  bool warnNoSupport = false;
  bool verbose = false;

  filename[0] = 0;
  UParam::ignoreLogicalOnlyCellsForSynthesis(false);
  /* liberty_parser_debug =1; */
	
  /* handle the usual arguments */

  /* if the list of possible arguments gets any longer, use the options package... */

  for (i=1; i< argc; i++) {
    if( !strcmp(argv[i],"--version") )
      version=1;
    else if( !strcmp(argv[i], "--verbose") )
      verbose=true;
    else if( !strcmp(argv[i], "--mem_verbose") )
      mempool_quiet=0;
    else if( !strcmp(argv[i], "--warn_no_support") )
      warnNoSupport = true;
    else if( !strcmp(argv[i], "--help") )
      help=1;
    else if( i == argc-1  && argv[i][0] != '-' )
      strcpy(filename,argv[i]);
    else
      printf("Unrecognized argument '%s' ignored. Use --help for help.\n", argv[i]);
  }
	
  if (!mempool_quiet) {
    UMem::verbose();
  }
  if( help ) {
    usage();
    exit(0);
  }
	
  if( version ) {
    print_version();
    exit(0);
  }

  if( !filename[0] ) {
#if 0
    printf("No filename specified!\n\n");
    usage();
#endif
    runtests();      // run builtin tests if no filename specified
  } else {
    readLiberty(filename, warnNoSupport, verbose);
  }
  exit(0);
}
